#include <stdio.h>

#include <app/Application.h>
#include <app/Looper.h>
#include <interface/Window.h>
#include <interface/View.h>
#include "reflector/StorageReflectConstants.h"
#include "besupport/MessageTransceiverThread.h"
#include "util/PortableHashtable.h"
#include "util/PortableString.h"

// This view handles the drawing of the little balls/game pieces
class DrawView : public BView
{
public:
   DrawView(BRect rect) : BView(rect, NULL, B_FOLLOW_ALL_SIDES, B_WILL_DRAW),
        _mtt(NULL), _whereTable(PortableStringHashFunc), _localPoint(-100,-100)
   {
      SetViewColor(100,100,100);
   }

   virtual void MouseMoved(BPoint where, uint32 code, const BMessage * m)
   {
      BView::MouseMoved(where, code, m);
      _localPoint = where;

      // Now tell the server our new mouse position
      if (_mtt) 
      {
         // This message represents our new position
         PortableMessage * whereMsg = new PortableMessage('mpos');
         whereMsg->AddPoint("where", PortablePoint(where.x, where.y));
         PortableMessageRef whereRef(whereMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());

         // This message is the command to the server to update its database
         // Specifically, it should update our demogame/where node with the
         // new position
         PortableMessage * setDataMsg = new PortableMessage(PR_COMMAND_SETDATA);
         setDataMsg->AddMessage("demogame/where", whereRef);
         _mtt->SendOutgoingMessage(PortableMessageRef(setDataMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool()));
      }
      Invalidate();
   }
 
   virtual void Draw(BRect ur)
   {
      if (_mtt)
      {
         PortableHashtableIterator<PortableString, BPoint> iter = _whereTable.GetIterator();
         BPoint pt;
         while(iter.GetNextValue(pt))
         {
            SetHighColor(0,255,0);
            FillEllipse(pt, 5, 5);
         }

         SetHighColor(255,0,0);
         FillEllipse(_localPoint, 5, 5);
      }
   }

   MessageTransceiverThread * _mtt;  // not owned by us!  (owned by the window)
   PortableHashtable<PortableString, BPoint> _whereTable;  // table mapping sessionID to ball location
   BPoint _localPoint;  // where the local player's ball is (near the mouse pointer usually)
};

// This BWindow subclass is used to process the incoming
// BMessages that our MessageTransceiverThread sends to it
class GameWindow : public BWindow
{
public:
   GameWindow(const char * serverName)
      : BWindow(BRect(100,100,500,500), "Silly MUSCLE game thingy", B_TITLED_WINDOW, B_NOT_RESIZABLE|B_NOT_ZOOMABLE)
   {
      _serverName = serverName;  // "beshare.bentonrea.com" or whatnot
      _mtt = NULL; // just in case

      _view = new DrawView(Bounds());
      AddChild(_view);
   }

   ~GameWindow()
   {
      delete _mtt;  // shut down and get rid of the MessageTransceiverThread
   }
   
   virtual bool QuitRequested()
   {
      be_app->PostMessage(B_QUIT_REQUESTED);
      return true;
   }

   virtual thread_id Run()
   {
      thread_id tid = BLooper::Run();
      if (tid >= 0)
      {
         BMessage connectMsg('conn');     // message we want to see when the connection is made
         BMessage disconnectMsg('disc');  // message we want to see when the connection is broken

         // Startup!  Create our MessageTransceiverThread... that is our
         // interface to MUSCLE and to the Internet.
         _mtt = new MessageTransceiverThread(BMessenger(this));  // have it send replies back to me
         if (_mtt->StartConnectThread(_serverName.Cstr(), 2960, &connectMsg, &disconnectMsg) == B_NO_ERROR)
         {
            printf("Now connecting to the MUSCLE server...\n");
         }
         else
         {
            printf("Aggh!  StartConnectThread() failed... net_server needs restarting???\n");
            PostMessage(B_QUIT_REQUESTED);
         }
      }
      return tid;
   }

   virtual void MessageReceived(BMessage * msg)
   {
      switch(msg->what)
      {
         case 'conn':
         {
            printf("MUSCLE says the TCP connection to the MUSCLE server has been completed/connected!!\n");
            _view->SetViewColor(200,200,200);
            _view->Invalidate();
            _view->_mtt = _mtt;  // so our view can use it too

            {
               // Now that we are connected, upload a node to the server to tell everyone
               // that we are a demogame client (and not, say, a BeShare client)
               PortableMessage * data = new PortableMessage;  // this could be any message, but we'll leave it empty
               PortableMessageRef dataMsgRef(data, NULL, NULL);

               // This is the actual command to the server to set an ID node for us
               PortableMessage * setCommand = new PortableMessage(PR_COMMAND_SETDATA);
               setCommand->AddMessage("demogame", dataMsgRef);
               PortableMessageRef setCommandRef(setCommand, NULL, NULL);
               _mtt->SendOutgoingMessage(setCommandRef);
            }

            {
               // Lastly, let's "subscribe" to the demogame node and the where node
               // This way, the muscle daemon will tell us whenever another
               // demogame node is added or removed from his database;
               // that lets us know if another demogame program has
               // logged in or logged out.
               PortableMessage * subscribeMessage = new PortableMessage(PR_COMMAND_SETPARAMETERS);
               subscribeMessage->AddBool("SUBSCRIBE:demogame/where", true);
               PortableMessageRef subscribeRef(subscribeMessage, NULL, NULL);
               _mtt->SendOutgoingMessage(subscribeRef);
            }
         }
         break;

         case 'disc':
            printf("MUSCLE says the TCP connection  to the MUSCLE serverhas been broken/disconnected!!\n");
            _view->SetViewColor(100,100,100);
            _view->Invalidate();
            _view->_mtt = NULL;  // this will keep our view from trying to send when there is no point
         break;

         case PORTABLE_MESSAGES_RECEIVED:
         {
            bool changed = false;
            PortableMessageRef msgRef;
            while(_mtt->GetNextIncomingMessage(msgRef) == B_NO_ERROR)
            {
               PortableMessage * pmsg = msgRef.GetItemPointer();
               if (pmsg->what == PR_RESULT_DATAITEMS)
               {
                  // First, see if there are any notifications that any "demogame/where" nodes 
                  // have been removed.  If so, remove the corresponding ball from the playing field
                  const char * np;
                  for (int i=0; (pmsg->FindString(PR_NAME_REMOVED_DATAITEMS, i, &np) == B_NO_ERROR); i++)
                  {
                     PortableString sessionID;
                     if (GetSessionIDFromWherePath(np, sessionID) == B_NO_ERROR)
                     {
                        _view->_whereTable.Remove(sessionID);
                        changed = true;
                     }
                  }

                  // Now look and see if there are any notifications that a "demogame/where" node
                  // has been added or updated.  If so, update our ball position for that guy.
                  PortableString nodepath;
                  PortableMessageFieldNameIterator iter = pmsg->GetFieldNameIterator(B_MESSAGE_TYPE);
                  while(iter.GetNextFieldName(nodepath))
                  {     
                     PortableString sessionID;
                     if (GetSessionIDFromWherePath(nodepath, sessionID) == B_NO_ERROR)
                     {
                        // Okay, the path matches... now extract the positional data from the message.
                        // We only need the last/latest one (in case there are several)
                        int32 count;
                        if (pmsg->GetInfo(nodepath.Cstr(), NULL, &count) == B_NO_ERROR)
                        {
                           PortableMessageRef whereRef;
                           if (pmsg->FindMessage(nodepath.Cstr(), count-1, whereRef) == B_NO_ERROR)
                           {
                              PortablePoint p;
                              if (whereRef.GetItemPointer()->FindPoint("where", p) == B_NO_ERROR)
                              {
                                 _view->_whereTable.Put(sessionID, BPoint(p.x, p.y));
                                 changed = true;
                              }
                           }
                        }
                     }
                  }
               }
            }
            _view->Invalidate();
         }
         break;

         default:
            BLooper::MessageReceived(msg);
         break;
      }
   }

   MessageTransceiverThread * GetMessageTransceiverThread() {return _mtt;}

private:
   // Utility function to get the session ID from a /ipaddress/sessionid/demogame/where path
   status_t GameWindow :: GetSessionIDFromWherePath(const PortableString & path, PortableString & retID) const
   {
      retID = path;

      // Expect a string like this:  /ipaddress/sessionid/demogame/where
      if (retID.EndsWith("/demogame/where"))
      {
         int secondSlash = retID.IndexOf('/', 1);
         if (secondSlash >= 0)
         {
            retID = retID.Substring(secondSlash+1);  // chop off the beginning...
            retID = retID.Substring(0, retID.Length()-15);
            return B_NO_ERROR;
         }
      }
      return B_ERROR;
   }

   PortableString _serverName;
   MessageTransceiverThread * _mtt;
   DrawView * _view;
};

void main(int argc, char ** argv)
{
   BApplication app("application/x-vnd.sugoi.demogame");

   // Create and start up our incoming-messages-processing looper
   GameWindow * myWindow = new GameWindow((argc > 1) ? argv[1] : "beshare.bentonrea.com");
   myWindow->Show();
 
   app.Run();

   if (myWindow->Lock()) myWindow->Quit();  // shut down the looper

   printf("Bye bye!\n"); 
   return 0;
}
