/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef NETWORK_UTILITY_FUNCTIONS_H
#define NETWORK_UTILITY_FUNCTIONS_H

#include "util/TimeUtilityFunctions.h"

// On these OS's, calls like accept() take an int* rather than a size_t*
#if __BEOS__ || __APPLE__ || __CYGWIN__
# define net_length_t int
#else
# define net_length_t size_t
#endif


/** Convenience method for connecting with TCP to a given hostName/port.
 * @param hostName The host name (or ASCII IP address) of the computer to connect to.
 * @param port The port number to connect to.
 * @param debugTitle If non-NULL, debug output to stdout will be enabled and debug output will be prefaced by this string.
 * @param debugOutputOnErrorsOnly if true, debug output will be printed only if an error condition occurs.
 * @return the sockfd if the connection is successful (greater than or equal to 0), -1 if the connection attempt failed.
 */
int Connect(const char * hostName, int port, const char * debugTitle = NULL, bool debugOutputOnErrorsOnly = true);

/** Returns a string representation of the IP address (address).  Useful on OS's where inet_ntoa isn't available.
 *  @param address The 4-byte IP address to translate into text.
 *  @param outBuf The ASCII representation of the IP address will be written into this buffer,
 *                which must be at least 13 bytes long.
 */
void Inet_NtoA(uint32 address, char * outBuf);

/** Fills out (retHostIP) and (retPort) with information about the remote peer on the other end of (socket).  
 *  @param socket The socket descriptor to find out info about.
 *  @param retHostIP On success, the IP address of the computer that (socket) is connected to
 *                   will be written into this object.
 *  @param retPort On success, the remote port that (socket) is connected to will be written into this object.
 *  @return B_NO_ERROR on success, B_ERROR on failure (such as if the socket isn't valid and connected).
 */
status_t GetPeerInfo(int socket, uint32 & retHostIP, uint16 & retPort);

/** Calls fork(), setsid(), chdir(), umask(), etc, to fork an independent daemon process.
 *  Also closes all open file descriptors.
 *  Note that this function will call exit() on the parent process if successful,
 *  and thus won't ever return in that process. 
 *  @param optNewDir If specified, the daemon will chdir() to the directory specified here.
 *  @param optOutputTo Where to redirect stderr and stdout to.  Defaults to "dev/null".
 *                     If set to NULL, or if the output device can't be opened, output
 *                     will not be rerouted.
 *  @param createOutputFileIfNecessary if set true, and (optOutputTo) can't be opened,
 *                                     (optOutputTo) will be created.
 *  @return B_NO_ERROR on success (the child process will see this), B_ERROR on failure.
 */
status_t BecomeDaemonProcess(const char * optNewDir = NULL, const char * optOutputTo = "/dev/null", bool createOutputFileIfNecessary = false);

/** Creates a pair of sockets that are connected to each other,
 *  so that any bytes you pass into one socket come out the other socket.
 *  This is useful when you want to wake up a thread that is blocked in select()...
 *  you have it select() on one socket, and you send a byte on the other.
 *  @param retSocket1 On success, this value will be set to the socket ID of the first socket.  Set to -1 on failure.
 *  @param retSocket2 On success, this value will be set to the socket ID of the second socket.  Set to -1 on failure.
 *  @param blocking Whether the two sockets should use blocking I/O or not.  Defaults to false.
 *  @return B_NO_ERROR on success, or B_ERROR on failure.
 */
status_t CreateConnectedSocketPair(int & retSocket1, int & retSocket2, bool blocking = false);

/** Enables or disables blocking I/O on the given socket. 
 *  (Default for a socket is blocking mode enabled)
 *  @param socket the socket descriptor to act on.
 *  @param enabled Whether I/O on this socket should be enabled or not.
 *  @return B_NO_ERROR on success, or B_ERROR on failure.
 */
status_t SetSocketBlockingEnabled(int socket, bool enabled);

#endif

