/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef PORTABLE_REFCOUNT_H
#define PORTABLE_REFCOUNT_H

#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"
#include "util/ObjectPool.h"

// Multithread safety is provided by two macros, ATOMIC_PREINCREMENT and ATOMIC_PREDECREMENT.
// For PortableRefs to work safely in multithreaded environments, these macros
// must be defined to increment or decrement their argument ATOMICALLY and return
// the resulting value.  Note that these macros are defined correctly for BeOS
// only... for other systems they are defined as the regular (non-atomic) match
// operators.  So if you are porting to another OS and are using multithreaded
// code, you will need to #define these to the appropriate functions of your OS.
#ifdef __BEOS__
# define ATOMIC_PREINCREMENT(c) (atomic_add(&c,1)+1)
# define ATOMIC_PREDECREMENT(c) (atomic_add(&c,-1)-1)
#else
# define ATOMIC_PREINCREMENT(c) (++c)  // not thread safe!!!
# define ATOMIC_PREDECREMENT(c) (--c)  // not thread safe!!!
#endif

/**
 *  This class implements reference counting on any object or array
 *  you give it, so that your object will be automatically deleted when 
 *  your last reference to it goes away.  Less error-prone than doing the deletes
 *  by hand, and faster than a full garbage collector; the only thing to watch
 *  out for is circular dependencies which could keep your references from ever
 *  going away...
 */
template <class Item> class PortableRef
{
public:
   class RefCountMem;
   typedef ObjectPool<RefCountMem> RefPool;  /**< type of an ObjectPool of our internal data structures */
   typedef ObjectPool<Item> ItemPool;        /**< type of an ObjectPool of user data structures */

   /** 
    * Creates a new reference-count for the given item.
    * Once this reference count has been created, (item) will be automatically
    * deleted when the last PortableRef that references it has dissappeared.
    * Note that it is an error to call this constructor twice with the same (item).
    * To obtain additional tracked references to (item), use PortableRef's copy 
    * constructor, or the PortableRef equals operator.
    *
    * (refPool) may be NULL, or if you wish to be more efficient, it may point to a RefPool
    * that you are using to recycle ref-count memory for this PortableRef class.  (Note that
    * the calling code retains ownership of the RefPool object)
    *
    * (itemPool) may be NULL, or if you wish to be more efficient, it may point to a ItemPool
    * that you are using to recycle your (item)s.  (Note that
    * the calling code retains ownership of the ItemPool object)
    *
    * If (item) is a pointer to a single item, and not an array of items, then
    * the default value of one is appropriate.
    *
    * (numItems) is the number of (Item) objects that (item) points to.  In general,
    * this will be one, but for arrays it might be more.
    *
    * If (doRefCounting) is set false, then (item) will NOT be deleted automatically.
    * It then remains the responsibility of the calling code to delete it.  (This is useful
    * if e.g. you need to put a non-new'd item into a PortableRef, and want to 
    * handle the cleanup issues yourself)
    *
    * Usage example: 
    *
    * PortableRef<Foo> ref(new Foo());
    * for (int i=0; i<10; i++) _vectors[i].AddItem(ref);
    * (the Foo object will be deleted automagically at the right time, later...)
    *
    * @param item A dynamically allocated object that the PortableRef class will assume responsibility for deleting.
    * @param itemPool An optional ObjectPool to recycle the object into.  May be NULL.
    * @param refPool An optional ObjectPool to recycle our internal ref-count data into.  May be NULL.
    * @param numItems Number of Items pointed to by (item).  Usually 1, but may be different if (item)
    *                 is actually pointing to an array of Items.
    * @param doRefCounting If set false, the PortableRef will never delete the Item.  
    *                      (this is only useful in very rare cases)  If set to false, the calling
    *                      code will retain responsibility for deleting Item at the proper time.
    */
   PortableRef(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems = 1, bool doRefCounting = true)
   {
      if (doRefCounting)
      {
         _rc = (refPool) ? refPool->GetObject() : newnothrow RefCountMem();
         if (_rc) 
         {
            _rc->Init(item, itemPool, refPool, numItems);
            _defaultPtr = NULL;
            _ownedItemCount = -1;  // _rc owns the pointer now
         }
         else
         {
            // can't allocate refCount!  Fall back to just holding it ourself (sole ownership)
            WARN_OUT_OF_MEMORY;
            _defaultPtr = item;
            _ownedItemCount = numItems;  // we own the pointer
         }
      }
      else
      {
         _rc = NULL;
         _defaultPtr = item;
         _ownedItemCount = -1;  // calling code still owns the pointer here
      }
   }

   /** Copy constructor.  Creates an additional reference to the object referenced by (copyMe).
    *  The referenced object won't be deleted until ALL PortableRefs that reference it are gone.
    */
   PortableRef(PortableRef & copyMe) : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1)
   {
      *this = copyMe; 
   }

   /** Default constructor.
    *  Creates a NULL reference (suitable for initializing with the = operator)
    */
   PortableRef() : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1) 
   {
      // empty
   }

   /** Unreferences the held data item.  If this is the last PortableRef that
    *  references the held data item, the data item will be deleted or recycled at this time.
    */
   ~PortableRef() 
   {
      Reset();
   }

   /** Assigment operator.
    *  Unreferences the previous held data item, and adds a reference to the the data item of (r).
    */
   PortableRef &operator=(const PortableRef &r) 
   {
      if (&r != this)
      {
         DeleteOwnedDefaultPointer();
         if (_rc) _rc->Unref();
         _rc = r._rc;
         if (_rc) _rc->Ref();
         _defaultPtr = (r._ownedItemCount >= 0) ? NULL : r._defaultPtr;  // if he is sole owner, we don't want it
      }
      return *this;
   }

   /** Returns true iff both PortableRefs are referencing the same data. */
   int operator ==(const PortableRef &rhs) const {return _rc == rhs._rc;}
 
   /** Returns the ref-counted data item.  The returned data item
    *  is only guaranteed valid for as long as this PortableRefCount object exists.
    */
   Item * GetItemPointer() const {return _rc ? _rc->GetItem() : _defaultPtr;}

   /** Convenience synonym for GetItemPointer(). */
   Item * operator()() const {return GetItemPointer();}

   /** Returns the number of items in the array we are tracking (or
    *  one if we are tracking just a single object)
    */
   uint32 GetItemCount() const {return _rc ? _rc->GetItemCount() : 0L;}

   /** Convenience method:  Unreferences our held data item (if any), and 
    *  turns this object back into a NULL reference.  (equivalent to *this = PortableRef();)
    */
   void Reset() 
   {
      if (_rc) _rc->Unref();
      _rc = NULL;
      DeleteOwnedDefaultPointer();
   }
   
private:
   void DeleteOwnedDefaultPointer()
   {
      if (_ownedItemCount >= 0) 
      {
         if (_ownedItemCount == 1) delete _defaultPtr;
                               else delete [] _defaultPtr;
         _ownedItemCount = -1;
         _defaultPtr = NULL;
      }
   }

   /** This class is used internally by the PortableRef class.  An object of this class
    *  represents the reference count for a given object, and is shared by all the PortableRef
    *  classes referencing the tracked object.
    */
   class RefCountMem
   { 
   public:
      RefCountMem() {/* empty */}

      void Init(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems)
      {
         _refCount  = 1;
         _item      = item;
         _itemPool  = itemPool;
         _refPool   = refPool;
         _numItems  = numItems;
      }

      ~RefCountMem() 
      {
         FreeObject();  // shouldn't be necessary but just in case
      }

      void FreeObject()
      {
         if (_item)
         {
            if (_itemPool) _itemPool->ReleaseObject(_item, _numItems);
            else 
            {
               if (_numItems == 1) delete _item;
                              else delete [] _item;
            }
            _item = NULL;
         }
      }
  
      void Ref()  
      {
         (void) ATOMIC_PREINCREMENT(_refCount); 
      }

      void Unref() 
      {
         if (ATOMIC_PREDECREMENT(_refCount) == 0) 
         {
            FreeObject();
            if (_refPool) _refPool->ReleaseObject(this);
                     else delete this;
         }
      }

      Item * GetItem() const {return _item;}

   private:
      Item * _item;
      int32 _refCount;
      RefPool * _refPool;
      ItemPool * _itemPool;
      uint32 _numItems;
   };

   RefCountMem * _rc;
   Item * _defaultPtr;  // used only when user is handling object deletion himself
   int _ownedItemCount;  // if >=0, we have sole ownership of (_defaultPtr), and _ownedItemCount is the array size
};

#endif
